"""
TODO
"""

import setproctitle
import logging
from Xlib import X, display
from .helper.helper_ewmh import ewmh
from .main import regularize
from .util_kdtree import insert_focused_window_into_kdtree
from .windowlist import windowlist


def search_window(win, lst):
    lst = list(lst)
    logging.debug('search window')
    for win_target, win_test in lst:
        logging.debug('test window %s', win_test.id)
        if win_test.id == win.id:
            logging.debug('find window:%s', win_target.id)
            return win_target

    plst = []
    for win_target, win_test in lst:
        pwin = win_test.query_tree().parent
        if pwin:
            logging.debug('window %s has parent %s', win_test.id, pwin.id)
            plst.append((win_target, pwin))
        else:
            logging.debug('window %s has no parent', win_test.id)
    if len(plst) > 0:
        return search_window(win, plst)
    else:
        logging.debug("can't find window:%s", win.id)
        return None


def loop():
    setproctitle.setproctitle("freetile-auto")

    disp = display.Display()
    root = disp.screen().root
    # | X.SubstructureRedirectMask)
    root.change_attributes(event_mask=X.SubstructureNotifyMask)

    wininfo = {}

    IGNORE_TYPES = [
        disp.intern_atom('_NET_WM_WINDOW_TYPE_DESKTOP'),
        disp.intern_atom('_NET_WM_WINDOW_TYPE_DOCK'),
        disp.intern_atom('_NET_WM_WINDOW_TYPE_TOOLBAR'),
        disp.intern_atom('_NET_WM_WINDOW_TYPE_MENU'),
        disp.intern_atom('_NET_WM_WINDOW_TYPE_UTILITY'),
        disp.intern_atom('_NET_WM_WINDOW_TYPE_SPLASH'),
        disp.intern_atom('_NET_WM_WINDOW_TYPE_DIALOG'),
    ]
    IGNORE_STATES = [
        disp.intern_atom('_NET_WM_STATE_ABOVE'),
        disp.intern_atom('_NET_WM_STATE_STICKY'),
        disp.intern_atom('_NET_WM_STATE_SKIP_TASKBAR'),
        disp.intern_atom('_NET_WM_STATE_SKIP_PAGER'),
    ]
    IGNORE_TYPES = set(IGNORE_TYPES)
    IGNORE_STATES = set(IGNORE_STATES)

    regularize()

    def insert_window(win):
        c = win.get_wm_class()
        n = win.get_wm_name()
        s = ewmh.getWmState(win)
        t = ewmh.getWmWindowType(win)

        if not IGNORE_STATES.isdisjoint(s):
            logging.debug('wmstate:%s', s)
            return False
        if len(t) < 1 or not IGNORE_TYPES.isdisjoint(t):
            logging.debug('window type:%s', t)
            return False
        if c is not None and 'Popup' in c:
            logging.debug('wmclass:%s', c)
            return False

        wininfo[win.id] = c, n, t, s
        return True

    for win in ewmh.getClientList():
        insert_window(win)

    def add_window(win):
        logging.debug('add window')
        for _ in range(50):
            try:
                ewmh.display.flush()
                ewmh.display.sync()
                lst = ewmh.getClientList()
                newwin = search_window(win, zip(lst, lst))
                if newwin:
                    break
            except BaseException:
                logging.info('fail %s' % _)
                return True

        win = newwin
        if win:
            logging.debug('window id:%s', win.id)
            if insert_window(win):
                logging.info([e.type, *wininfo[win.id]])
                windowlist.reset()
                # move new window to top.
                windowlist.windowInCurrentWorkspaceInStackingOrder.remove(win.id)
                windowlist.windowInCurrentWorkspaceInStackingOrder.append(win.id)
                num = len(windowlist.windowInCurrentWorkspaceInStackingOrder)
                if not regularize(ignore_overlapped_layout=True, minimum_regularized_window=num - 1):
                    return False
        return True

    while True:
        e = disp.next_event()
        if e.type == X.MapNotify:
            logging.debug('MapNotify event')
            win = e.window
            if not add_window(win):
                logging.info('failed to add new window')
                break
        elif e.type == X.UnmapNotify:
            logging.debug('UnmapNotify event')
            win = e.window
            if win.id in wininfo:
                logging.info([e.type, *wininfo.pop(win.id)])
                windowlist.reset(ignore=[win.id])
                if len(windowlist.windowInCurrentWorkspaceInStackingOrder) < 1:
                    logging.info('no window exists')
                    # quit auto tiling when no window exists.
                    break
                regularize()
        elif e.type == X.ConfigureNotify:
            logging.debug('ConfigureNotify event')
        elif e.type == X.ClientMessage:
            logging.debug('ClientMessag event')
        elif e.type == X.MappingNotify:
            logging.debug('MappingNotify event')
        elif e.type == X.DestroyNotify:
            logging.debug('DestroyNotify event')
        else:
            logging.debug('unknown event:%s', e.type)
    # Quit loop when detect overlapped windows created by user.
    logging.info('quit autotiling')



#!/usr/bin/python
window_gap = 4
left_padding = 4
bottom_padding = 4
top_padding = 4
right_padding = 4
navigate_across_workspaces = False    # True  # availabe in viewports desktops

exclude_wm_name = []
exclude_wm_class = []

resize_step = 72
move_step = 72
min_window_width = 50
min_window_height = 50

max_tree_branch = 3

fullscreen_tiling = True
maximize_single_window = True

# overwrite default configuration
import os.path

config_file = os.path.expanduser("~/.config/freetilerc")
try:
    s = open(config_file).read()
    exec(s)
except BaseException:
    pass



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import logging
import json

from .divide import divide


class Node:
    """
    Build up a 2d tree.
    """
    parent = None
    children = None
    key = None
    position = None
    """node resized by user"""
    resized = False
    DIMENSION = 2
    __leafnodemap = None

    def __str__(self):
        obj = self.to_json()
        return json.dumps(obj, indent=4, sort_keys=True)

    def to_json(self):
        return {
            '1.key': self.key,
            '2.position': self.position,
            '3.resized': self.resized,
            '4.children': [child.to_json() for child in self.children] if self.children else None,
        }

    def leaf(self):
        return self.children is None

    def leafnodemap(self):
        if self.parent is None:
            # root
            return self.__leafnodemap
        else:
            return self.parent.leafnodemap()

    def dimension(self):
        """
        The dimension expected to be controlled by current node and divided between children.
        """
        # column first
        index_min = self.depth() % self.DIMENSION
        index_max = self.depth() % self.DIMENSION + self.DIMENSION
        return index_min, index_max

    def targets(self):
        dmin, dmax = self.parent.dimension()
        if dmin == 0:
            return 'left', 'right'
        elif dmin == 1:
            return 'up', 'down'

    def interval_size(self):
        dmin, dmax = self.parent.dimension()
        return self.position[dmax] - self.position[dmin]

    def __init__(self, _input, parent=None):
        # create empty node
        if _input is None:
            return
        if parent is None:
            # root
            self.__leafnodemap = {}
        else:
            self.parent = parent

        if len(_input) == 1:
            pos, key = _input[0]
            self.key = key
            self.leafnodemap()[key] = self
        else:
            dmin, dmax = self.dimension()
            _input2 = [[pos[dmin], pos[dmax]] for pos, v in _input]
            intervals = divide(zip(_input2, _input))
            if len(intervals) == 1 and self.depth() > 0:
                pass
            else:
                self.children = []
                for item in intervals:
                    _list = [nodelist for interval, nodelist in item]
                    node_child = Node(_list, parent=self)
                    self.children.append(node_child)

        self.position = self.init_position(_input)

    def init_position(self, _input=None):
        if self.leaf():
            position_children = [position for position, value in _input]
        else:
            position_children = [child.position for child in self.children]
        minx, miny = 10**6, 10**6
        maxx, maxy = -10**6, -10**6
        minx = min([pos[0] for pos in position_children])
        miny = min([pos[1] for pos in position_children])
        maxx = max([pos[2] for pos in position_children])
        maxy = max([pos[3] for pos in position_children])
        return [
            min(10**6, minx),
            min(10**6, miny),
            max(-10**6, maxx),
            max(-10**6, maxy),
        ]

    def depth(self):
        if self.parent is None:
            return 0
        return 1 + self.parent.depth()

    def overlap(self):
        if self.leaf():
            return self.key is None

        for child in self.children:
            if child.overlap():
                return True
        return False

    def create_parent(self):
        new_parent = Node(None)
        new_parent.parent = self.parent
        new_parent.children = [self]
        new_parent.position = list(self.position)

        if self.parent is not None:
            index = self.parent.children.index(self)
            self.parent.children[index] = new_parent
        else:
            # root
            new_parent.__leafnodemap = self.__leafnodemap

        self.parent = new_parent

        return new_parent

    def create_sibling(self):
        sibling = Node(None)
        sibling.parent = self.parent
        sibling.position = list(self.position)

        index = self.parent.children.index(self)
        self.parent.children.insert(index + 1, sibling)
        return sibling

    def children_resized(self, gap):
        dmin, dmax = self.dimension()
        start = self.position[dmin]
        end = self.position[dmax]
        gap = gap[dmin]
        size = len(self.children)
        interval = int((end - start + gap) / size) - gap
        i = start
        for index in range(size):
            child = self.children[index]
            min_expect = i
            max_expect = i + interval
            i += interval + gap
            if index == size - 1:
                max_expect = end
            cmin = child.position[dmin]
            cmax = child.position[dmax]
            if not cmin == min_expect or not cmax == max_expect:
                return True
        return False

    def regularize(self, gap):
        """
        Regularize a branch in a 2d tree. A node's position is a rectangle
        defined by x0,y0,x1,y1. All the children's positions will be adjusted
        to fit exactly to their parents.
        """
        if self.leaf():
            return
        dmin, dmax = self.dimension()

        for child in self.children:
            for i in range(2 * self.DIMENSION):
                if i not in [dmin, dmax]:
                    child.position[i] = self.position[i]
        i0 = self.position[dmin]
        i1 = self.position[dmax]
        b = gap[dmin]
        num = len(self.children)

        resizeed_index = -1
        for i in range(num):
            child = self.children[i]
            if child.resized:
                resizeed_index = i

        i0 = i0 - b
        size = i1 - i0 - num * b
        size_sum = 0
        if resizeed_index < 0:
            for child in self.children:
                size_sum += child.interval_size()
            i = i0
            default_interval_size = int(size / num)
            for child in self.children:
                i += b
                if size == size_sum:
                    _size = child.interval_size()
                else:
                    _size = default_interval_size
                    # _size += int(child.interval_size() * size / size_sum)
                child.position[dmin] = i
                i += _size
                child.position[dmax] = i
        else:
            for i in range(num):
                child = self.children[i]
                if i > resizeed_index \
                        or 1 + resizeed_index == num \
                        and not child.resized:
                    size_sum += child.interval_size()
                else:
                    size -= child.interval_size()

            i = i0
            for index in range(num):
                child = self.children[index]
                i += b
                if size == size_sum:
                    _size = child.interval_size()
                else:
                    _size = child.interval_size()
                    if index > resizeed_index or resizeed_index + \
                            1 == num and not child.resized:
                        _size = int(_size * size / size_sum)
                child.position[dmin] = i
                i += _size
                child.position[dmax] = i

        self.children[-1].position[dmax] = self.position[dmax]
        for child, i in zip(self.children, range(len(self.children))):
            child.regularize(gap)

    def remove_from_tree(self):
        """
        Remove a node from a 2d tree, and keep children.
        """
        if self.children is None:
            return
        l = []
        if len(self.children) == 1:
            child = self.children[0]
            if child.children is not None:
                if self.parent is not None:
                    i = self.parent.children.index(self)
                    self.parent.children.remove(self)
                    for grandchild in child.children:
                        self.parent.children.insert(i, grandchild)
                        grandchild.parent = self.parent
                        i += 1
                l = child.children
        else:
            l = self.children
        for child in l:
            child.remove_from_tree()

    def getLayout(self, size_limit_map):
        """
        Extract layouts and window ids from a 2d tree.
        """
        if self.children is None:
            x0, y0, x1, y1 = self.position
            minw, minh = size_limit_map[self.key]
            layout = [x0, y0, x1 - x0, y1 - y0]
            return [layout], [self.key], x1 - x0 < minw or y1 - y0 < minh
        else:
            layouts, values, reach_size_limit = [], [], False
            for child in self.children:
                l, v, r, = child.getLayout(size_limit_map)
                layouts += l
                values += v
                reach_size_limit = reach_size_limit or r
            return layouts, values, reach_size_limit


if __name__ == '__main__':
    print('test')



"""
"""
import logging
from .windowlist import windowlist
from .workarea import workarea


def main():
    logging.info('Start test')
    logging.debug('load windows info')
    windowlist.reset()

    logging.debug('current work area:%s,%s,%s,%s', workarea.x, workarea.y, workarea.width, workarea.height)

    winid = windowlist.windowInCurrentWorkspaceInStackingOrder[-1]
    logging.debug('current active window id:%s', winid)
    geo = windowlist.windowGeometry[winid]
    logging.debug('current active window geo:%s', geo)
    target = [workarea.width // 4, workarea.height // 4, workarea.width // 2, workarea.height // 2]
    logging.info("Try to move current window to %d,%d and resize to %dx%d", *target)
    windowlist.arrange([target], [winid])
    logging.debug('sleep 1s')
    import time
    time.sleep(1)
    logging.debug('load windows info')
    windowlist.reset()
    winid = windowlist.windowInCurrentWorkspaceInStackingOrder[-1]
    logging.debug('current active window id:%s', winid)
    newgeo = windowlist.windowGeometry[winid]
    logging.info('Current window is moved to %d,%d and resized to %dx%d', *geo)
    if newgeo == target:
        logging.info('Test succeeded')
    else:
        logging.debug('Test failed')



from . import config
from . import nontree

from .util_kdtree import (find_kdtree, move_kdtree, regularize_or_insert_windows, resize_kdtree)
from .windowlist import windowlist
from .workarea import workarea

import logging
logger = logging.getLogger(__name__)


def regularize(ignore_overlapped_layout=False, minimum_regularized_window=None):
    '''
    Try to regularize windows or add a new window into the K-D tree.
    '''
    logger.info('regularize layout')
    stack = windowlist.windowInCurrentWorkspaceInStackingOrder
    num = len(stack)
    if num == 0:
        pass
    elif num == 1 and config.maximize_single_window:
        windowlist.maximize_window(stack[0])
    else:
        """
        Allow to insert 1/3 new windows to a regularized layout of the rest 2/3 windows.
        """
        if minimum_regularized_window is None:
            minimum_regularized_window = max(min(num - 1, 2), num * 2 // 3, 1)

        if regularize_or_insert_windows(minimum_regularized_window):
            logger.info('regularize windows')
        elif not ignore_overlapped_layout:
            logger.info('force tiling')
            force_tiling()
        else:
            return False

        # Make sure the current active window is raised to top of the stack.
        # The stack order will be used to find out previous active windows.
        active = windowlist.get_active_window()
        if active is not None:
            windowlist.raise_window(active)

    return True


def force_tiling():
    winlist = windowlist.windowInCurrentWorkspaceInStackingOrder
    tile = workarea.tile(len(winlist))
    return windowlist.arrange(tile, winlist)


def resize(resize_width, resize_height):
    if resize_kdtree(resize_width, resize_height):
        return True
    logger.debug('failed to resize as a tree layout')
    if nontree.moveandresize([0, 0, resize_width, resize_height]):
        return True


def move(target):
    return move_kdtree(target, allow_create_new_node=True) \
        or move_kdtree(target, allow_create_new_node=False) \
        or nontree.move(target)


def swap(target):

    active = windowlist.get_active_window()

    if active is None:
        return False

    target_window_id = find_kdtree(active, target, False) \
        or nontree.find(active, target) \
        or find_kdtree(active, target, True)

    if target_window_id is None:
        from .monitor import monitor
        if monitor.available:
            target_window_id = nontree.find(active, target, True)

    if target_window_id is None:
        return False

    lay0 = windowlist.windowGeometry[active]
    lay1 = windowlist.windowGeometry[target_window_id]

    windowlist.arrange([lay0, lay1], [target_window_id, active])
    return True


def focus(target):

    active = windowlist.get_active_window(allow_outofworkspace=False)

    target_window_id = find_kdtree(active, target, False) \
        or nontree.find(active, target, config.navigate_across_workspaces) \
        or find_kdtree(active, target, True)

    if target_window_id is None:
        from .monitor import monitor
        if monitor.available:
            target_window_id = nontree.find(active, target, True)

    if target_window_id is None:
        windowlist.raise_window(active)
    else:
        windowlist.raise_window(target_window_id)
    return True


def grow_width():
    resize(config.resize_step, 0)


def grow_height():
    resize(0, config.resize_step)


def shrink_width():
    resize(-config.resize_step, 0)


def shrink_height():
    resize(0, -config.resize_step)



from .config import (
    bottom_padding,
    left_padding,
    right_padding,
    top_padding,
    window_gap,
)
from .helper.helper_ewmh import ewmh

import subprocess
import re


class Monitor:
    x = None
    y = None
    width = None
    height = None
    available = False

    def findMonitor(self, xywh0):
        """
        find current monitor
        """
        s = subprocess.check_output('xrandr').decode()
        l = re.findall('(\d+)x(\d+)\+(\d+)\+(\d+)', s)
        monitors = [(int(x), int(y), int(w), int(h)) for w, h, x, y in l]
        if len(monitors) > 0:
            window = ewmh.getActiveWindow()
            if window:
                win = window.get_geometry()
                for item in monitors:
                    self.x, self.y, self.width, self.height = item
                    if xywh0 is None:
                        return (0, 0, self.width, self.height)
                    x0, y0, w0, h0 = xywh0

                    xin = win.x < x0 + self.x + self.width and win.x + win.width > x0 + self.x
                    yin = win.y < y0 + self.y + self.width and win.y + win.width > y0 + self.y
                    if xin and yin:
                        self.available = True and len(monitors) > 2
                        if (self.x < 1 and self.y < 1):
                            return self.x + x0, self.y + y0, w0, h0
                        else:
                            return self.x + x0, self.y + y0, \
                            min(self.width, w0 + x0 - self.x), \
                            min(self.height, h0 + y0 - self.y)
        self.available = False
        return x0, y0, w0, h0


monitor = Monitor()



from .config import (exclude_wm_name, exclude_wm_class, min_window_height, min_window_width)

from .helper import xcb
import logging
from .helper.helper_ewmh import (ewmh, get_window_list, maximize_window, raise_window, unmaximize_windows)
from .helper.xlib import disp, get_frame_extents, get_wm_class_and_state
from .workarea import workarea


class WindowList:
    windowInCurrentWorkspaceInStackingOrder = []
    windowGeometry = {}
    windowName = {}
    minGeometry = {}
    windowObjectMap = {}
    ewmhactive = None

    def maximize_window(self, winid):
        win = self.windowObjectMap[winid]
        geo = win.get_geometry()
        if not workarea.windowInCurrentViewport(geo):
            from .monitor import monitor
            if monitor.available:
                xcb.move(winid, monitor.x, monitor.y)
            else:
                xcb.move(winid, 0, 0)
        logging.info('maximize window')
        maximize_window(win, sync=False)

    def raise_window(self, winid):
        logging.debug('raise window %s', winid)
        win = self.windowObjectMap[winid]
        raise_window(win)

    def reset(self, ignore=[]):
        # get_root_window_property("_NET_CURRENT_DESKTOP")
        desktop = ewmh.getCurrentDesktop()
        self.ewmhactive = ewmh.getActiveWindow()
        self.windowInCurrentWorkspaceInStackingOrder = []
        self.windowName = {}
        self.windowGeometry = {}
        self.minGeometry = {}
        self.windowObjectMap = {}

        for win, _desktop, name in get_window_list(ignore):
            winid = win.id
            self.windowObjectMap[winid] = win
            if not _desktop == desktop:
                continue
            if name in exclude_wm_name:
                continue

            if not {
                    disp.intern_atom('_NET_WM_STATE_SKIP_TASKBAR'),
            }.isdisjoint(ewmh.getWmState(win)):
                continue

            if not {disp.intern_atom('_NET_WM_WINDOW_TYPE_DOCK')}.isdisjoint(ewmh.getWmWindowType(win)):
                continue

            wmclass, minimized = get_wm_class_and_state(win)
            if minimized:
                continue
            wmclass = set(wmclass)
            if not wmclass == wmclass - set(exclude_wm_class):
                continue

            self.windowName[winid] = name

            geo = win.get_geometry()
            self.get_absolute_geo(win)
            # window in current viewport?
            if not workarea.windowInCurrentViewport(geo, threshold=0.1):
                continue

            wnh = win.get_wm_normal_hints()
            f_left, f_right, f_top, f_bottom = get_frame_extents(win)
            minw = max(min_window_width, wnh.min_width + f_left, f_right)
            minh = max(min_window_height, wnh.min_height + f_top, f_right)
            self.minGeometry[winid] = minw, minh

            self.windowInCurrentWorkspaceInStackingOrder.append(winid)

    def get_absolute_geo(self, win):
        if win.id not in self.windowGeometry:
            geo = win.get_geometry()
            f_left, f_right, f_top, f_bottom = get_frame_extents(win)
            self.windowGeometry[win.id] = [geo.x - f_left, geo.y - f_top, geo.width + f_left + f_right, geo.height + f_top + f_bottom]
            p = win.query_tree().parent
            if p:
                pgeo = self.get_absolute_geo(p)
                self.windowGeometry[win.id][0] += pgeo[0]
                self.windowGeometry[win.id][1] += pgeo[1]
        return self.windowGeometry[win.id]

    def get_current_layout(self):
        return [self.windowGeometry[_id] for _id in self.windowInCurrentWorkspaceInStackingOrder]

    def get_active_window(self, allow_outofworkspace=False):
        active = self.ewmhactive
        if active is None:
            return None
        active = active.id
        if active in self.windowInCurrentWorkspaceInStackingOrder:
            return active
        if allow_outofworkspace and active in self.windowName:
            return active

    def arrange(self, layout, windowids):
        windows = [self.windowObjectMap[wid] for wid in windowids]
        # unmaximize
        unmaximize_windows(windows)

        layout_final = []
        # move windows
        for win, lay, in zip(
                windows,
                layout,
        ):
            normal_hints = win.get_wm_normal_hints()

            x, y, width, height = lay
            f_left, f_right, f_top, f_bottom = get_frame_extents(win)
            width -= f_left + f_right
            height -= f_top + f_bottom

            # window gravity: Static
            if normal_hints.win_gravity == 10:
                y += f_top
                x += f_left
            layout_final.append([x, y, width, height])
        return xcb.arrange(layout_final, windowids)


windowlist = WindowList()



from .config import max_tree_branch, fullscreen_tiling, window_gap

from .tree import Node
from .windowlist import windowlist
from .workarea import workarea

import logging
logger = logging.getLogger(__name__)


def resize_kdtree(resize_width, resize_height):
    '''
    Adjust non-overlapping layout.
    '''

    winlist = windowlist.windowInCurrentWorkspaceInStackingOrder
    if len(winlist) < 2:
        logger.debug('ignore layouts with less than 2 windows')
        return False

    active = windowlist.get_active_window()
    if active is None:
        logger.debug('cannot find target window')
        return False

    lay = windowlist.get_current_layout()
    # generate k-d tree
    _tree = getkdtree(winlist, lay)
    if active not in _tree.leafnodemap():
        logger.debug('not a tree layout')
        return False
    current_node = _tree.leafnodemap()[active]

    regularize_node = None

    # resize nodes
    for node in [current_node, current_node.parent]:
        if node and node.parent and not node.overlap():
            index_min, index_max = node.parent.dimension()
            _resize = [resize_width, resize_height][index_min]
            if not _resize == 0:
                node.resized = True
                regularize_node = node.parent
                # invert the operation if the node is the last child of its
                # parent
                if regularize_node.children[-1] is node:
                    node.position[index_max] -= _resize
                else:
                    node.position[index_max] += _resize

    if regularize_node is None:
        logger.debug('cannot find target node')
        return True
    regularize_node = regularize_node.parent

    if fullscreen_tiling:
        _tree.position = [workarea.x, workarea.y, workarea.x + workarea.width, workarea.y + workarea.height]
        return regularize_kd_tree(_tree, ignore_size_limit_error=True)
    return regularize_kd_tree(regularize_node, ignore_size_limit_error=True)


def getkdtree(winlist, lay):
    origin_lay = [[x, y, x + w, y + h] for x, y, w, h in lay]
    lst = list(zip(origin_lay, winlist))
    return Node(lst)


def insert_window_into_kdtree(winid, target):
    winlist = [w for w in windowlist.windowInCurrentWorkspaceInStackingOrder if not w == winid]
    lay = windowlist.get_current_layout()
    _tree = getkdtree(winlist, lay)
    if target not in _tree.leafnodemap():
        # overlapped
        return False
    target_node = _tree.leafnodemap()[target]
    if target_node.parent is None:
        # If there is only one window node and it is root.
        _tree = target_node.create_parent()
    elif target_node.parent.children_resized(gap=(window_gap, window_gap)):
        # If the node is resized by user, don't resize it in the same axis
        # again.
        target_node.create_parent()

    node = target_node.create_sibling()
    node.key = winid
    if fullscreen_tiling:
        _tree.position = [workarea.x, workarea.y, workarea.x + workarea.width, workarea.y + workarea.height]
        return regularize_kd_tree(_tree, ignore_size_limit_error=True)
    return regularize_kd_tree(node.parent, ignore_size_limit_error=True)


def move_kdtree(target, allow_create_new_node=True):
    '''
    Adjust non-overlapping layout.
    '''

    active = windowlist.get_active_window()
    # can find target window
    if active is None:
        return False

    # sort_win_list(windowlist.windowInCurrentWorkspaceInStackingOrder)
    winlist = windowlist.windowInCurrentWorkspaceInStackingOrder
    # ignore layouts with less than 2 windows
    if len(winlist) < 2:
        return False

    lay = windowlist.get_current_layout()
    # generate k-d tree
    _tree = getkdtree(winlist, lay).create_parent().create_parent()
    if active not in _tree.leafnodemap():
        return False
    current_node = _tree.leafnodemap()[active]

    shift = 0 if target in ['left', 'up'] else 1

    # whether promote node to its parent's level
    if target not in current_node.targets():
        current_node.parent.children.remove(current_node)
        regularize_node = current_node.parent.parent
        if regularize_node is None:
            # If parent node is root.
            _tree = _tree.create_parent().create_parent()
            regularize_node = current_node.parent.parent
        index_parent = regularize_node.children.index(current_node.parent)
        regularize_node.children.insert(index_parent + shift, current_node)
    else:
        regularize_node = current_node.parent
        index_current = regularize_node.children.index(current_node)
        regularize_node.children.remove(current_node)
        '''
        If there is no more nodes at the target direction, promote the current node.
        '''
        if 0 <= index_current - 1 + shift < len(regularize_node.children):

            # If there are is only one sibling node, promote them both.
            if len(regularize_node.children) == 1:
                #
                shift = -1 if target in ['left', 'up'] else 1
                regularize_node.children.insert(index_current + shift, current_node)
            else:
                new_parent = regularize_node.children[index_current - 1 + shift]
                # If there is a leaf node at the target direction, build a new
                # parent node for the leaf node and the current node.
                _swap = False or not allow_create_new_node
                if new_parent.leaf() and not _swap:
                    # But allow no more than one branch for each node
                    non_leaf_node_count = 0
                    for sibling in new_parent.parent.children:
                        if not sibling.leaf() or max_tree_branch < 1:
                            non_leaf_node_count += 1
                            if not non_leaf_node_count < max_tree_branch:
                                # Just swap them.
                                _swap = True
                                break
                    else:
                        new_parent = new_parent.create_parent()
                if _swap:
                    shift = -1 if target in ['left', 'up'] else 1
                    regularize_node.children.insert(index_current + shift, current_node)
                else:
                    new_parent.children.append(current_node)
        else:
            # promote the current node.
            regularize_node = current_node.parent.parent.parent
            index_current = regularize_node.children.index(current_node.parent.parent)
            regularize_node.children.insert(index_current + shift, current_node)

        if len(regularize_node.children) == 1:
            regularize_node = regularize_node.parent

    # remove nodes which has only one child
    regularize_node.remove_from_tree()

    if regularize_node.overlap():
        return False

    # regularize k-d tree
    regularize_node = regularize_node.parent
    if fullscreen_tiling:
        _tree.position = [workarea.x, workarea.y, workarea.x + workarea.width, workarea.y + workarea.height]
        return regularize_kd_tree(_tree, ignore_size_limit_error=True)
    return regularize_kd_tree(regularize_node, ignore_size_limit_error=True)


def search_for_regularized_windows(_min, _max, stack, layout_stack):
    """
    search for regularized windows from stack bottom
    """
    result = None
    while _min <= _max:
        num = int(_min / 2 + _max / 2)
        winlist = stack[:num]
        layout = layout_stack[:num]
        tree = getkdtree(winlist, layout)
        if tree.overlap():
            _max = num - 1
        else:
            result = tree, winlist, num
            _min = num + 1
    return result


def regularize_or_insert_windows(min_regularized_window):
    stack = windowlist.windowInCurrentWorkspaceInStackingOrder
    layout_stack = windowlist.get_current_layout()
    num = len(stack)

    result = search_for_regularized_windows(min_regularized_window, num, stack, layout_stack)
    if result is None:
        return False
    tree, winlist, num = result
    logger.debug('original tree:%s', tree)

    target = winlist[-1]
    target_node = tree.leafnodemap()[target]

    tree.regularize(gap=(window_gap, window_gap))

    if target_node.parent is None:
        # If there is only one window node and it is root.
        tree = target_node.create_parent()
    elif target_node.parent.children_resized(gap=(window_gap, window_gap)):
        # if node is resized by user, dont resize it in the same axis
        # again.
        target_node.create_parent()
    # add all the rest windows to tree.
    for winid in stack[num:]:
        node = target_node.create_sibling()
        node.key = winid
    logger.debug('new tree:%s', tree)

    if fullscreen_tiling:
        tree.position = [workarea.x, workarea.y, workarea.x + workarea.width, workarea.y + workarea.height]
    tree.regularize(gap=(window_gap, window_gap))
    logger.debug('regularized new tree:%s', tree)
    # load k-d tree
    a, b, reach_size_limit = tree.getLayout(windowlist.minGeometry)
    if reach_size_limit:
        logger.info('reach window minimal size')
        return True
    windowlist.arrange(a, b)
    return True


def regularize_kd_tree(regularize_node, ignore_size_limit_error=False):
    if regularize_node is None:
        return False
    if regularize_node.overlap():
        return False
    # regularize k-d tree
    regularize_node.regularize(gap=(window_gap, window_gap))

    # load k-d tree
    a, b, reach_size_limit = regularize_node.getLayout(windowlist.minGeometry)
    if reach_size_limit:
        logger.info('reach window minimal size')
        if ignore_size_limit_error:
            return True
        return False
    windowlist.arrange(a, b)
    return True


def insert_focused_window_into_kdtree(newwin=None):
    if newwin is None:
        active = windowlist.get_active_window()
    else:
        active = newwin
    if active is None:
        return False
    last_active = get_last_active_window()
    if last_active is None:
        return False
    return insert_window_into_kdtree(active, last_active)


def get_last_active_window():
    active = windowlist.get_active_window()
    lst = windowlist.windowInCurrentWorkspaceInStackingOrder
    index = lst.index(active)
    index = (index - 1) % len(lst)
    return lst[index]


def find_kdtree(center, target, allow_parent_sibling=True):
    '''
    Adjust non-overlapping layout.
    '''
    active = center

    if active is None:
        return None

    winlist = windowlist.windowInCurrentWorkspaceInStackingOrder

    if active not in winlist:
        return None
    lay = windowlist.get_current_layout()
    _tree = getkdtree(winlist, lay)
    if active not in _tree.leafnodemap():
        return None
    current_node = _tree.leafnodemap()[active]

    if len(lay) == 1:
        return None

    promote = target not in current_node.targets()

    shift = -1 if target in ['left', 'up'] else 1

    promoted = False

    c = current_node
    if promote:
        c = c.parent
        promoted = True

    while True:
        i = c.parent.children.index(c)
        if 0 <= i + shift < len(c.parent.children):
            target_node = c.parent.children[i + shift]
            break
        if c.parent.parent is None or c.parent.parent.parent is None:
            return None
        c = c.parent.parent
        promoted = True

    if promoted:
        if not allow_parent_sibling:
            if not target_node.leaf():
                return None
    if target_node is None or target_node.overlap():
        return None

    return find_kdtree_get_leaf(target_node, target)


def find_kdtree_get_leaf(node, target):
    # is leaf
    if node.key:
        return node.key
    # is not leaf
    if target not in node.targets() and target in ['up', 'left']:
        node = node.children[-1]
    else:
        node = node.children[0]
    return find_kdtree_get_leaf(node, target)



"""
Freely Tiling Script for X

Usage:
  freetile [options] tile
  freetile [options] (focus|move|swap) (up|down|left|right)
  freetile [options] (grow|shrink) (height|width)
  freetile [options] autotile
  freetile [options] test
  freetile -h | --help

Options:
  -h --help     Show this screen.
  --debug       Debug
"""
import logging
from docopt import docopt
from freetile.main import swap, move, regularize, focus, grow_width, grow_height, shrink_width, shrink_height
from freetile.windowlist import windowlist


def main():
    arguments = docopt(__doc__)

    if arguments['--debug']:
        logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')
    else:
        logging.basicConfig(level=logging.INFO)

    for target in ('up', 'down', 'left', 'right'):
        if arguments[target]:
            break

    windowlist.reset()
    if arguments['swap']:
        swap(target)
    elif arguments['move']:
        move(target)
    elif arguments['focus']:
        focus(target)
    elif arguments['tile']:
        regularize()
    elif arguments['grow']:
        if arguments['width']:
            grow_width()
        else:
            grow_height()
    elif arguments['shrink']:
        if arguments['width']:
            shrink_width()
        else:
            shrink_height()
    elif arguments['autotile']:
        from freetile.auto import loop
        loop()
    elif arguments['test']:
        from freetile.test import main
        main()


if '__main__' == __name__:
    main()



from ewmh import EWMH
from Xlib import X

ewmh = EWMH()

_NET_WM_STATE_MAXIMIZED_VERT = ewmh.display.get_atom('_NET_WM_STATE_MAXIMIZED_VERT')
_NET_WM_STATE_MAXIMIZED_HORZ = ewmh.display.get_atom('_NET_WM_STATE_MAXIMIZED_HORZ')
_NET_WM_STATE_FULLSCREEN = ewmh.display.get_atom('_NET_WM_STATE_FULLSCREEN')


def raise_window(win):
    ewmh.setActiveWindow(win)
    ewmh.display.flush()


def maximize_window(win, sync=True, flush=True):
    if not sync:
        ewmh.setWmState(win, 1, '_NET_WM_STATE_MAXIMIZED_VERT', '_NET_WM_STATE_MAXIMIZED_HORZ')
        if flush:
            ewmh.display.flush()
    else:
        wmstate = set(ewmh.getWmState(win))
        if len({_NET_WM_STATE_MAXIMIZED_HORZ, _NET_WM_STATE_MAXIMIZED_VERT} - wmstate) < 1:
            return
        win.change_attributes(event_mask=X.StructureNotifyMask)
        ewmh.setWmState(win, 1, '_NET_WM_STATE_MAXIMIZED_VERT', '_NET_WM_STATE_MAXIMIZED_HORZ')
        ewmh.display.flush()
        timeout = 10
        for _ in range(timeout):
            e = ewmh.display.next_event()
            if e.type == X.ConfigureNotify:
                if e.window.id == win.id:
                    break
        win.change_attributes(event_mask=0)


def unmaximize_windows(winlist):
    """
    synchronized
    """
    # find maximized windows
    maximized_windows = set()
    for win in winlist:
        wmstate = ewmh.getWmState(win)
        if not {_NET_WM_STATE_MAXIMIZED_HORZ, _NET_WM_STATE_MAXIMIZED_VERT, _NET_WM_STATE_FULLSCREEN}.isdisjoint(wmstate):
            maximized_windows.add(win)

    # enable and send event
    for win in maximized_windows:
        win.change_attributes(event_mask=X.StructureNotifyMask)
        ewmh.setWmState(
            win,
            0,
            '_NET_WM_STATE_MAXIMIZED_VERT',
            '_NET_WM_STATE_MAXIMIZED_HORZ',
        )
        ewmh.setWmState(
            win,
            0,
            '_NET_WM_STATE_FULLSCREEN',
        )

    # flush
    ewmh.display.flush()

    # synchronize
    unmaximized_windows = set()
    timeout = len(maximized_windows) * 10
    for _ in range(timeout):
        if not len(maximized_windows) > len(unmaximized_windows):
            break
        e = ewmh.display.next_event()
        if e.type == X.ConfigureNotify:
            unmaximized_windows.add(e.window.id)

    # disable event
    for win in maximized_windows:
        win.change_attributes(event_mask=0)


def get_window_list(ignore=[]):
    """
    in stacking order.
    """
    for win in ewmh.getClientListStacking():
        if win.id in ignore:
            continue
        try:
            name = ewmh.getWmName(win)
            d=ewmh.getWmDesktop(win)
        except:
            name=None
            d=None
        if name is not None:
            name = name.decode('utf8')
        yield win, d, name



"""
Operations for overlapped layout.
"""
from .windowlist import windowlist
from .workarea import workarea
from . import config


def moveandresize(target):
    active = windowlist.get_active_window(allow_outofworkspace=True)
    # cannot find target window
    if active is None:
        return False
    lay = windowlist.windowGeometry[active]
    for i in range(4):
        lay[i] += target[i]
    windowlist.arrange([lay], [active])
    return True


def move(target):
    target = {
        'left': [-config.move_step, 0, 0, 0],
        'down': [0, config.move_step, 0, 0],
        'up': [0, -config.move_step, 0, 0],
        'right': [config.move_step, 0, 0, 0],
    }[target]
    return moveandresize(target)


def cal_center(x, y, w, h):
    return [x + w / 2.2, y + h / 2.2]


def find(center, target, allow_outofworkspace=False):
    '''
    find the nearest window in the target direction.
    '''

    if center is None:
        lay_center = workarea.width / 2.0, workarea.height / 2.0
    else:
        lay_center = windowlist.windowGeometry[center]
        lay_center = cal_center(*lay_center)
    _min = -1
    _r = None
    if allow_outofworkspace:
        winlist = windowlist.windowName
    else:
        winlist = windowlist.windowInCurrentWorkspaceInStackingOrder
    for w in winlist:
        l = windowlist.windowGeometry[w]
        l = cal_center(*l)
        bias1, bias2 = 1.0, 1.0
        bias = 4.0
        if target == 'down':
            delta = l[1] - lay_center[1]
            delta2 = (l[1] - lay_center[1])**2 - (l[0] - lay_center[0])**2
            bias1 = bias
        if target == 'up':
            delta = lay_center[1] - l[1]
            delta2 = (l[1] - lay_center[1])**2 - (l[0] - lay_center[0])**2
            bias1 = bias
        if target == 'right':
            delta = l[0] - lay_center[0]
            delta2 = (l[0] - lay_center[0])**2 - (l[1] - lay_center[1])**2
            bias2 = bias
        if target == 'left':
            delta = lay_center[0] - l[0]
            delta2 = (l[0] - lay_center[0])**2 - (l[1] - lay_center[1])**2
            bias2 = bias
        distance = bias1 * (l[0] - lay_center[0]) ** 2 + \
            bias2 * (l[1] - lay_center[1]) ** 2
        if delta > 0 and delta2 > 0:
            if _min == -1 or distance < _min:
                _min = distance
                _r = w
    return _r



#!/usr/bin/env python
# -*- coding: UTF-8 -*-


def divide(_input):
    '''
    divide intervals
    '''
    # find out separable intervals
    _input = sorted([n for n in _input])
    l = [interval for interval, value in _input]
    for i0, i1 in l:
        assert not i0 == i1
    current_start, current_end = None, None
    result = []
    for start, end in l:    # zip(l[:-1], l[1:]):
        if current_start is None:
            current_start, current_end = start, end
        else:
            if start >= current_end:
                result.append((current_start, current_end))
                current_start, current_end = start, end
            else:
                current_end = max(end, current_end)
    result.append((current_start, current_end))

    nodes = result
    # assign children to intervals
    children = []
    l = []
    current_start, current_end = nodes.pop(0)
    for child in _input:
        interval, value = child
        start, end = interval
        if not end <= current_end:
            children.append(l)
            current_start, current_end = nodes.pop(0)
            l = []
        l.append(child)
    children.append(l)

    return children


if __name__ == '__main__':
    _input = [[0.3, 0.4], [0.2, 0.5], [0.6, 0.8], [0.8, 0.9], [0.8, 1.0]]
    _input = list(zip(_input, range(len(_input))))
    # print(_input)
    _out = [[([0.2, 0.5], 1), ([0.3, 0.4], 0)], [([0.6, 0.8], 2)], [([0.8, 0.9], 3), ([0.8, 1.0], 4)]]
    try:
        assert divide(_input) == _out
    except BaseException:
        print(divide(_input))

    _input = [[-0.1, 0.4], [0.3, 0.4], [0.2, 0.5], [0.6, 0.8], [0.8, 0.9], [0.8, 1.0]]
    _input = list(zip(_input, range(len(_input))))
    _out = [[
        ([-0.1, 0.4], 0),
        ([0.2, 0.5], 2),
        ([0.3, 0.4], 1),
    ], [([0.6, 0.8], 3)], [([0.8, 0.9], 4), ([0.8, 1.0], 5)]]

    try:
        assert divide(_input) == _out
    except BaseException:
        print(divide(_input))
    # print(divide(_input))
    # print(_input)
    _input = [[-0.1, 0.4], [1.1, 1.2], [0.3, 0.4], [0.2, 0.5], [0.6, 0.8], [0.8, 0.9], [0.8, 1.0]]
    _input = list(zip(_input, range(len(_input))))
    # print(_input)
    print(divide(_input))



import Xlib
from Xlib import X, display, protocol

disp = display.Display()
screen = disp.screen()
root = screen.root


def edit_prop(window, mode, name, value):
    cm_event = protocol.event.ClientMessage(window=window, client_type=disp.intern_atom(name), data=(32, [mode, disp.intern_atom(value), 0, 0, 0]))
    disp.send_event(root, cm_event, (X.SubstructureRedirectMask | X.SubstructureNotifyMask))


def get_root_window_property(name):
    return root.get_property(disp.intern_atom(name), Xlib.Xatom.CARDINAL, 0, 32).value


def get_frame_extents(win):
    frame_extents = win.get_property(disp.intern_atom("_NET_FRAME_EXTENTS"), Xlib.Xatom.CARDINAL, 0, 32)
    if frame_extents is None:
        return 0, 0, 0, 0
    else:
        return frame_extents.value


def get_wm_class_and_state(win):
    wm_class = win.get_wm_class()
    wm_state = win.get_wm_state()
    minimized = not wm_state.state == 1
    return wm_class, minimized



import math
from .config import (
    bottom_padding,
    left_padding,
    right_padding,
    top_padding,
    window_gap,
)

from .helper.helper_ewmh import ewmh
from .monitor import monitor


class WorkArea:
    x = None
    y = None
    width = None
    height = None

    def __init__(self):
        xywh = ewmh.getWorkArea()
        xywh = None if xywh is None else xywh[:4]
        x, y, w, h = monitor.findMonitor(xywh)
        self.width = w - left_padding - right_padding
        self.height = h - top_padding - bottom_padding
        self.x = x + left_padding
        self.y = y + top_padding

    def tile(self, wincount):
        # for rotated screen
        cols = 1 if workarea.width < workarea.height else 2
        layout = []
        colwidth = int((self.width + window_gap) / cols) - window_gap
        windowsleft = wincount
        x = self.x
        for col in range(cols):
            if col == cols - 1:
                colwidth = self.width + self.x - x
            rows = min(int(math.ceil(float(wincount) / cols)), windowsleft)
            windowsleft -= rows
            rowheight = int((window_gap + self.height) / rows) - window_gap
            y = self.y
            for row in range(rows):
                if row == rows - 1:
                    rowheight = self.height + self.y - y
                layout.append([x, y, colwidth, rowheight])
                y += rowheight + window_gap
            x += colwidth + window_gap

        return layout[:wincount]

    def windowInCurrentViewport(self, geo, threshold=1 / 2):
        _threshold = geo.width * threshold
        if _threshold > self.width + self.x - geo.x:
            return False
        if _threshold > geo.x + geo.width - self.x:
            return False
        _threshold = geo.height * threshold
        if _threshold > self.height + self.y - geo.y:
            return False
        if _threshold > geo.y + geo.height - self.y:
            return False
        return True


workarea = WorkArea()



from setuptools import setup
import glob

setup(
    name='freetile',
    version='v0.1.3',
    description='freely tiling script for X',
    url='http://github.com/rbn42/freetile',
    download_url='https://github.com/rbn42/freetile/archive/v0.1.3.tar.gz',
    author='rbn42',
    author_email='bl100@students.waikato.ac.nz',
    license='MIT',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: X11 Applications',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Desktop Environment :: Window Managers',
    ],
    keywords=['window', 'tiling', 'xorg'],
    packages=['freetile', 'freetile.helper'],
    install_requires=[
        'docopt',
        'ewmh',
        'xcffib',
        'python-xlib',
        'setproctitle',
    ],
    data_files=[
        ('share/doc/visualbar/config', glob.glob('doc/config/*')),
    ],
    entry_points={
        'console_scripts': [
            'freetile=freetile.__main__:main',
        ]
    })



import os

import xcffib
from xcffib.testing import XvfbTest
from xcffib.xproto import Atom, ConfigWindow, EventMask, GetPropertyType

conn = xcffib.connect(os.environ['DISPLAY'])
xproto = xcffib.xproto.xprotoExtension(conn)


def arrange(layout, windowids):
    for lay, winid in zip(layout, windowids):
        xproto.ConfigureWindow(winid, ConfigWindow.X | ConfigWindow.Y | ConfigWindow.Width | ConfigWindow.Height, lay)
    conn.flush()


def move(winid, x, y, sync=True):
    xproto.ConfigureWindow(winid, ConfigWindow.X | ConfigWindow.Y, [x, y])
    if sync:
        conn.flush()
